/*
 * Network block device - make block devices work over TCP
 *
 * Note that you can not swap over this thing, yet. Seems to work but
 * deadlocks sometimes - you can not swap over TCP in general.
 * 
 * Copyright 1997 Pavel Machek <pavel@atrey.karlin.mff.cuni.cz>
 * 
 * (part of code stolen from loop.c)
 *
 * 97-3-25 compiled 0-th version, not yet tested it 
 *   (it did not work, BTW) (later that day) HEY! it works!
 *   (bit later) hmm, not that much... 2:00am next day:
 *   yes, it works, but it gives something like 50kB/sec
 * 97-4-01 complete rewrite to make it possible for many requests at 
 *   once to be processed
 * 97-4-11 Making protocol independent of endianity etc.
 * 97-9-13 Cosmetic changes
 * 98-5-13 Attempt to make 64-bit-clean on 64-bit machines
 * 99-1-11 Attempt to make 64-bit-clean on 32-bit machines <ankry@mif.pg.gda.pl>
 * 2000-01-03 A bunch of changes to the protocol to add support for
 *            gzipped data and versioning <asun@cobalt.com>. the gzip stuff
 *            is turned off by default as it doesn't seem to win
 *            in my test cases. 
 *
 * possible FIXME: make set_sock / set_blksize / set_size / do_it one syscall
 * why not: would need verify_area and friends, would share yet another 
 *          structure with userland
 */

#define PARANOIA
#include <linux/major.h>

#include <linux/module.h>

#include <linux/sched.h>
#include <linux/fs.h>
#include <linux/malloc.h>
#include <linux/stat.h>
#include <linux/errno.h>
#include <linux/file.h>
#include <linux/ioctl.h>

#include <asm/segment.h>
#include <asm/uaccess.h>
#include <asm/types.h>

#define MAJOR_NR NBD_MAJOR
#include <linux/nbd.h>

#define USE_GZIP 0

#if USE_GZIP
/* mondo yucko. zlib.c should be in the toplevel lib directory. */
#include "../net/zlib.c"
#undef next
#endif

#define LO_MAGIC 0x68797548

static int nbd_blksizes[MAX_NBD];
static int nbd_blksize_bits[MAX_NBD];
static int nbd_sizes[MAX_NBD];
static u64 nbd_bytesizes[MAX_NBD];

static struct nbd_device nbd_dev[MAX_NBD];

#if 1
#define DEBUG( s )
#else
#define DEBUG( s ) printk( s ) 
#endif

#ifdef PARANOIA
static int requests_in;
static int requests_out;
#endif

#ifndef MIN
#define MIN(a, b) ((a) < (b) ? (a) : (b))
#endif

#define NBD_TIMEOUT (HZ*2)

#if USE_GZIP
/* zlib.c should have a default kernel zalloc/zfree */
static void *zalloc(void *arg, unsigned int items, unsigned int size)
{
	return kmalloc(items*size, GFP_KERNEL);
}

static void zfree(void *arg, void *ptr)
{
	kfree(ptr);
}
#endif

static int nbd_open(struct inode *inode, struct file *file)
{
	int dev;
	struct nbd_device *nbdev;

	if (!inode)
		return -EINVAL;
	dev = MINOR(inode->i_rdev);
	if (dev >= MAX_NBD)
		return -ENODEV;

	nbdev = &nbd_dev[dev];
	nbd_dev[dev].refcnt++;
	if (!(nbdev->flags & NBD_INITIALISED)) {
		nbdev->queue_lock = MUTEX;
		nbdev->flags |= NBD_INITIALISED;
	}
	MOD_INC_USE_COUNT;
	return 0;
}

/*
 * Send or receive packet. to coalesce packets, we stick a header [+
 * data] into an iovec. NOTE: if you just want to send a header, it's
 * better to use the second buffer.  */
static int nbd_xmit(const int send, struct nbd_device *lo, 
		    void *header, unsigned int hsize,
		    void *buf, unsigned int size)
{
	mm_segment_t oldfs;
	int result;
	struct msghdr msg;
	struct iovec iov[2];
	unsigned int total;
	unsigned long flags, timeout;
	sigset_t oldset;

	oldfs = get_fs();
	set_fs(get_ds());

	spin_lock_irqsave(&current->sigmask_lock, flags);
	oldset = current->blocked;
	sigfillset(&current->blocked);
	recalc_sigpending(current);
	spin_unlock_irqrestore(&current->sigmask_lock, flags);

	/* NOTE: we must be in non-blocking mode and have 
	 * signals turned off or else we expose kernel space to user
	 * access. */
	total = hsize + size;
	if (lo->timeout)
		timeout = jiffies + lo->timeout;
	while (1) {
		memset(&msg, 0, sizeof(msg));
		msg.msg_flags = MSG_NOSIGNAL | MSG_DONTWAIT;
		msg.msg_iov = iov;
		if (hsize) {
			iov[0].iov_base = header;
			iov[0].iov_len = hsize; 
			iov[1].iov_base = buf;
			iov[1].iov_len = size;
			msg.msg_iovlen = 2;
		} else {
			iov[0].iov_base = buf;
			iov[0].iov_len = size;
			msg.msg_iovlen = 1;
		}

		result = (send) ? sock_sendmsg(lo->sock, &msg, total) : 
			sock_recvmsg(lo->sock, &msg, total, MSG_DONTWAIT);

		if ((result == -EAGAIN) && (!lo->timeout || 
		    ((long) (timeout - jiffies) > 0))) { /* redo */ 
			set_fs(oldfs);
			schedule();
			oldfs = get_fs();
			set_fs(get_ds());
			continue;
		}

		if (result <= 0) {
#ifdef PARANOIA
			printk(KERN_ERR "NBD: %s - sock=%p at header=%p, hsize=%d, buf=%p, size=%d (total=%d) (result=%d).\n",
			       send ? "send" : "receive", lo->sock, 
			       header, hsize, buf, size, total, result);
#endif
			break;
		}

		total -= result;
		if (!total)
			break; /* done */

		/* fix up our buffers for another pass */
		if (hsize > result) {
			header += result;
			hsize -= result;
		} else {
			if (hsize) {
				result -= hsize;
				hsize = 0;
			}
			buf += result;
			size -= result;
		}
	} 
	
	spin_lock_irqsave(&current->sigmask_lock, flags);
	current->blocked = oldset;
	recalc_sigpending(current);
	spin_unlock_irqrestore(&current->sigmask_lock, flags);
	set_fs(oldfs);

	return result;
}

#define FAIL( s ) { printk( KERN_ERR "NBD: " s "(result %d)\n", result ); goto error_out; }

static void nbd_send_req(struct nbd_device *lo, const u32 cmd, const u64 from, 
			 void *buffer, const u32 len, struct request *req)
{
	int result;
	struct nbd_request request;

	DEBUG("NBD: sending control, ");
	request.magic = htonl(NBD_REQUEST_MAGIC);
	request.type = htonl(cmd);
	request.from = cpu_to_be64(from);
	request.len = htonl(len);
	memcpy(request.handle, &req, sizeof(req));

	if (len && !((cmd == NBD_OP_READ) || (cmd == NBD_OP_READZ))) {
		DEBUG("data, ");
		result = nbd_xmit(1, lo, &request, sizeof(request),
				  buffer, len);
	} else 
		result = nbd_xmit(1, lo, NULL, 0, &request, sizeof(request));
	if (result <= 0)
		FAIL("Send Control/Data failed.");
	DEBUG("done.\n");
	return;

 error_out:
	req->errors++;
}


/* we got a request. right now, it can only be a tickle. 
 * we need to prevent ourselves from getting interrupted here. */
static int nbd_handle_request(struct nbd_device *lo, void *handle)
{
	struct nbd_request *request = handle;
	int result;

	if (!request)
		return 0;

	/* suck in a request if desired */
	if (request->len) {
		if (request->len > lo->zlen)
			return -1;

		result = nbd_xmit(0, lo, NULL, 0, lo->zbuf, 
				  request->len);
		if (result <= 0)
			return result;
	}

	/* we only deal with tickles for now */
	if (request->type == ntohl(NBD_OP_TICKLE)) {
	  struct nbd_reply reply;

	  memset(&reply, 0, sizeof(reply));
	  reply.magic = htonl(NBD_REPLY_MAGIC);
	  reply.type = request->type;
	  nbd_xmit(1, lo, NULL, 0, &reply, lo->replysize);
	  DEBUG("sent tickle reply\n");
	}
	return 0;
}


#define HARDFAIL( s ) { printk( KERN_ERR "NBD: " s "(result %d)\n", result ); lo->harderror = result; return NULL; }
static struct request *nbd_read_stat(struct nbd_device *lo, 
				     struct request *handle)
		/* NULL returned = something went wrong, inform userspace       */ 
{
	int result, magic;
	struct nbd_reply reply;
	struct request *xreq, *req;
#if USE_GZIP
	z_streamp z;
#endif

read_stat_start:	
	DEBUG("reading control, ");
	reply.magic = 0;
	result = nbd_xmit(0, lo, NULL, 0, &reply, lo->replysize);
	if (result <= 0) {
		HARDFAIL("Recv control failed.");
	}

	/* this is only possible if we're using versioned nbd structs.
	 * NOTE: we can't return from nbd_read_stat if we get a 
	 * request. */
	magic = ntohl(reply.magic);
	if ((magic == NBD_REQUEST_MAGIC) && lo->s_version) {
	        DEBUG("got request\n");
		if (nbd_handle_request(lo, &reply) < 0)
			HARDFAIL("Invalid request!");
		/* we can't actually return to nbd_do_it as the 
		 * wrong thing will happen */
		goto read_stat_start;
	}
	  
	if (magic != NBD_REPLY_MAGIC)
		HARDFAIL("Not enough magic.");

	req = handle ? handle : lo->tail;
	memcpy(&xreq, reply.handle, sizeof(xreq));
	if (xreq != req)
		HARDFAIL("Unexpected handle received.\n");

	DEBUG("ok, ");
	if (ntohl(reply.error))
		FAIL("Other side returned error.");

	switch (lo->s_version ? ntohl(reply.type) : req->cmd) {
	case NBD_OP_READ:
		DEBUG("data, ");
		result = nbd_xmit(0, lo, NULL, 0, 
				  req->buffer, req->current_nr_sectors << 9);
		if (result <= 0)
			HARDFAIL("Recv data failed.");
		break;

#if USE_GZIP
	case NBD_OP_READZ:
		DEBUG("dataz, ");
		if (!lo->zbuf) 
			HARDFAIL("Unable to decompress data!");

		z = lo->istreamp;
		if ((z->avail_in = ntohl(reply.len)) > lo->zlen)
			HARDFAIL("Got an impossible reply size!");

		result = nbd_xmit(0, lo, NULL, 0, lo->zbuf, z->avail_in);
		if (result <= 0)
			HARDFAIL("Recv data failed.");

		z->next_in = lo->zbuf;
		z->next_out = req->buffer;
		z->avail_out = req->current_nr_sectors << 9;

		if ((result = inflate(z, Z_FINISH)) != Z_STREAM_END) 
			HARDFAIL("Recv uncompress failed.");
		inflateReset(z);
		break;
#endif
	}
	DEBUG("done.\n");
	return req;

/* Can we get here? Yes, if other side returns error */
      error_out:
	req->errors++;
	return req;
}

/* because we fake up our own request structure, 
 * we have to handle both the send and reply ourselves. */
static void nbd_send_config(struct nbd_device *lo)
{
	struct request req;
	u16 value;
	u8 buf[4]; /* config info */

	memset(buf, 0, sizeof(buf));

	/* fill up version and flags fields */
	value = htons(NBD_VERSION);
	memcpy(buf, &value, sizeof(value));

	/* tell the client that we support read compression. */
#if USE_GZIP
	if (lo->c_flags & NBD_FLAG_GZIP) {
	  value = htons(NBD_FLAG_GZIP);
	  memcpy(buf + sizeof(value), &value, sizeof(value));
	}
#endif

	DEBUG("NBD: sending config, ");
	req.cmd = NBD_OP_CONFIG;
	nbd_send_req(lo, NBD_OP_CONFIG, 0, buf, sizeof(buf), &req);
	nbd_read_stat(lo, &req);
	DEBUG("NBD: okay!\n");
}


/* initialize a few things */
static void nbd_pre_doit(struct nbd_device *lo, const int minor)
{
	/* if it's version 0, don't do anything */
	DEBUG("NBD: pre-doit\n");
	if (!lo->s_version) 
		return;

	/* first, set up our scratch buffer */
#if USE_GZIP
	if (lo->zbuf)
		kfree(lo->zbuf);
	lo->zlen = max_readahead[MAJOR_NR] ? max_readahead[MAJOR_NR][minor] :
	  MAX_READAHEAD;
	lo->zlen = MIN(lo->zlen, NBD_DATA_MAX);
	lo->zbuf = kmalloc(lo->zlen, GFP_KERNEL);
	if (!lo->zbuf) {
		printk(KERN_ALERT "NBD: Unable to allocate a scratch buffer for compression\n");
		goto pre_doit_err;
	}

	/* now, set up our input buffer. */
	lo->istreamp = kmalloc(sizeof(z_stream), GFP_KERNEL);
	if (lo->istreamp) {
		z_streamp z = lo->istreamp;

		/* set up memory allocation routines */
		memset(z, 0, sizeof(z_stream));
	 	z->zalloc = zalloc;
		z->zfree = zfree;
	
		if (inflateInit(z) != Z_OK) {
			kfree(lo->istreamp);
			lo->istreamp = NULL;
			printk(KERN_ALERT "NBD: Unable to initialize inflate buffer.\n");
			goto inflate_err;
		}
	} else {
		printk(KERN_ALERT "NBD: Unable to allocate space for compressed input.\n");
		goto inflate_err;
	}
	
	/* finally, set up our output buffer. if this fails, just turn
	 * off the compress flag and continue. */
	if (lo->s_flags & NBD_FLAG_GZIP) {
		lo->ostreamp = kmalloc(sizeof(z_stream), GFP_KERNEL);
		if (lo->ostreamp) {
			z_streamp z = lo->ostreamp;

			/* set up the memory allocation routines. */
			memset(z, 0, sizeof(z_stream));
			z->zalloc = zalloc;
			z->zfree = zfree;	
			if (deflateInit(z, Z_BEST_SPEED) == Z_OK) {
				return; /* okay */
			}
			kfree(lo->ostreamp);
			lo->ostreamp = NULL;
		}
		printk(KERN_ALERT "NBD: Unable to allocate space for compressed output\n");
		goto pre_doit_err;
	}

inflate_err:
	kfree(lo->zbuf);
	lo->zlen = 0;
	lo->zbuf = NULL;

pre_doit_err:
#endif
	lo->s_flags &= ~NBD_FLAG_GZIP;
}


/* de-initialize stuff */
static void nbd_post_doit(struct nbd_device *lo)
{
	DEBUG("NBD: post-doit\n");
	if (!lo->s_version)
		return;

#if USE_GZIP
	/* clean up compression bits */
	if (lo->istreamp) {
		inflateEnd(lo->istreamp);
		kfree(lo->istreamp);
		lo->istreamp = NULL;
	}
	
	if (lo->ostreamp) {
		deflateEnd(lo->ostreamp);
		kfree(lo->ostreamp);
		lo->ostreamp = NULL;
	}
	
	if (lo->zbuf) {
		kfree(lo->zbuf);
		lo->zlen = 0;
		lo->zbuf = NULL;
	}
#endif		
	/* now, reset version and flags */
	lo->s_version = lo->s_flags = 0;
        lo->replysize = NBD_REPLY_SIZE_V0;
}


static void nbd_do_it(struct nbd_device *lo)
{
	struct request *req;

	/* if the server reported a version, send off our info. */
	if (lo->s_version) 
		nbd_send_config(lo);

	while (1) {
		req = nbd_read_stat(lo, NULL);
		if (!req)
			return;
		down (&lo->queue_lock);
#ifdef PARANOIA
		if (req != lo->tail) {
			printk(KERN_ALERT "NBD: I have problem...\n");
		}
		if (lo != &nbd_dev[MINOR(req->rq_dev)]) {
			printk(KERN_ALERT "NBD: request corrupted!\n");
			goto next;
		}
		if (lo->magic != LO_MAGIC) {
			printk(KERN_ALERT "NBD: nbd_dev[] corrupted: Not enough magic\n");
			up (&lo->queue_lock);
			return;
		}
#endif
		nbd_end_request(req);
		if (lo->tail == lo->head) {
#ifdef PARANOIA
			if (lo->tail->next)
				printk(KERN_ERR "NBD: I did not expect this\n");
#endif
			lo->head = NULL;
		}
		lo->tail = lo->tail->next;
	next:
		up (&lo->queue_lock);
	}
}

void nbd_clear_que(struct nbd_device *lo)
{
	struct request *req;

	while (1) {
		req = lo->tail;
		if (!req)
			return;
#ifdef PARANOIA
		if (lo != &nbd_dev[MINOR(req->rq_dev)]) {
			printk(KERN_ALERT "NBD: request corrupted when clearing!\n");
			continue;
		}
		if (lo->magic != LO_MAGIC) {
			printk(KERN_ERR "NBD: nbd_dev[] corrupted: Not enough magic when clearing!\n");
			return;
		}
#endif
		req->errors++;
		nbd_end_request(req);
		if (lo->tail == lo->head) {
#ifdef PARANOIA
			if (lo->tail->next)
				printk(KERN_ERR "NBD: I did not assume this\n");
#endif
			lo->head = NULL;
		}
		lo->tail = lo->tail->next;
	}
}

/*
 * We always wait for result of write, for now. It would be nice to make it optional
 * in future
 * if ((req->cmd == WRITE) && (lo->flags & NBD_WRITE_NOCHK)) 
 *   { printk( "Warning: Ignoring result!\n"); nbd_end_request( req ); }
 */

#undef FAIL
#define FAIL( s ) { printk( KERN_ERR "NBD, minor %d: " s "\n", dev ); goto error_out; }

static void do_nbd_request(void)
{
	struct request *req;
	int dev;
	struct nbd_device *lo;

	while (CURRENT) {
		req = CURRENT;
		dev = MINOR(req->rq_dev);
#ifdef PARANOIA
		if (dev >= MAX_NBD)
			FAIL("Minor too big.");		/* Probably can not happen */
#endif
		lo = &nbd_dev[dev];
		if (!lo->file)
			FAIL("Request when not-ready.");
		if ((req->cmd == WRITE) && (lo->flags & NBD_READ_ONLY))
			FAIL("Write on read-only");
#ifdef PARANOIA
		if (lo->magic != LO_MAGIC)
			FAIL("nbd[] is not magical!");
		requests_in++;
#endif
		req->errors = 0;
		CURRENT = CURRENT->next;
		req->next = NULL;

		spin_unlock_irq(&io_request_lock);
		down (&lo->queue_lock);
		if (lo->head == NULL) {
			lo->head = req;
			lo->tail = req;
		} else {
			lo->head->next = req;
			lo->head = req;
		}

		/* if the other side supports compression,
		   try to compress the buffer on writes */
#if USE_GZIP
		if ((req->cmd == WRITE) &&
		    (lo->s_flags & NBD_FLAG_GZIP)) {
			z_streamp z = lo->ostreamp;
			int len = req->current_nr_sectors << 9;
			
			z->next_in = req->buffer;
			z->avail_in = len;
			z->next_out = lo->zbuf;
			z->avail_out = MIN(len, lo->zlen) - 1;

			/* we use z->avail_out to guarantee that a 
			 * successful deflation will be smaller than
			 * just sending the block. */
			if (deflate(z, Z_FINISH) == Z_STREAM_END) {
				nbd_send_req(lo, NBD_OP_WRITEZ,
					     (u64) req->sector << 9,
					     lo->zbuf, z->total_out, req);
				deflateReset(z);
				goto writez_ok;
			}
			deflateReset(z);
		}
#endif

		nbd_send_req(lo, req->cmd, (u64) req->sector << 9,
			     req->buffer, req->current_nr_sectors << 9,
			     req);	/* Why does this block?         */

#if USE_GZIP
writez_ok:
#endif
		up (&lo->queue_lock);
		spin_lock_irq(&io_request_lock);
		continue;

	      error_out:
		req->errors++;
		nbd_end_request(req);
		CURRENT = CURRENT->next;
	}
	return;
}

static int nbd_ioctl(struct inode *inode, struct file *file,
		     unsigned int cmd, unsigned long arg)
{
	struct nbd_device *lo;
	int dev, error, temp;

	/* Anyone capable of this syscall can do *real bad* things */

	if (!capable(CAP_SYS_ADMIN))
		return -EPERM;
	if (!inode)
		return -EINVAL;
	dev = MINOR(inode->i_rdev);
	if (dev >= MAX_NBD)
		return -ENODEV;

	lo = &nbd_dev[dev];
	switch (cmd) {
	case NBD_CLEAR_SOCK:
		nbd_clear_que(lo);
		if (lo->head || lo->tail) {
			printk(KERN_ERR "nbd: Some requests are in progress -> can not turn off.\n");
			return -EBUSY;
		}
		file = lo->file;
		if (!file)
			return -EINVAL;
		lo->file = NULL;
		lo->sock = NULL;
		fput(file);
		return 0;
	case NBD_SET_SOCK:
		if (lo->file)
			return -EBUSY;
		error = -EINVAL;
		file = fget(arg);
		if (file) {
			inode = file->f_dentry->d_inode;
			/* N.B. Should verify that it's a socket */
			lo->file = file;
			lo->sock = &inode->u.socket_i;
			error = 0;
		}
		return error;
	case NBD_SET_BLKSIZE:
		if ((arg & (arg-1)) || (arg < 512) || (arg > PAGE_SIZE))
			return -EINVAL;
		nbd_blksizes[dev] = arg;
		temp = arg >> 9;
		nbd_blksize_bits[dev] = 9;
		while (temp > 1) {
			nbd_blksize_bits[dev]++;
			temp >>= 1;
		}
		nbd_sizes[dev] = nbd_bytesizes[dev] >> nbd_blksize_bits[dev];
		nbd_bytesizes[dev] = nbd_sizes[dev] << nbd_blksize_bits[dev];
		return 0;
	case NBD_SET_SIZE:
		nbd_sizes[dev] = arg >> nbd_blksize_bits[dev];
		nbd_bytesizes[dev] = nbd_sizes[dev] << nbd_blksize_bits[dev];
		return 0;
	case NBD_SET_SIZE_BLOCKS:
		nbd_sizes[dev] = arg;
		nbd_bytesizes[dev] = ((u64) arg) << nbd_blksize_bits[dev];
		return 0;
	case NBD_SET_CLIENT_FLAGS:
	        if (arg & NBD_FLAG_GZIP) {
			lo->c_flags |= NBD_FLAG_GZIP;
		} else {
			lo->c_flags &= ~NBD_FLAG_GZIP;
		}
		return 0;
	case NBD_SET_SERVER_FLAGS:
	        if (arg & NBD_FLAG_GZIP) {
			lo->s_flags |= NBD_FLAG_GZIP;
		} else {
			lo->s_flags &= ~NBD_FLAG_GZIP;
		}
		return 0;
	case NBD_SET_SERVER_VERSION:
	        lo->s_version = arg;
		lo->replysize = (arg) ? NBD_REPLY_SIZE : NBD_REPLY_SIZE_V0;
		return 0;
	case NBD_SET_SERVER_TICKLE:
		/* convert from milliseconds to hz based */
		if (arg) 
			lo->timeout = (arg*HZ/1000)*2 + NBD_TIMEOUT;
		return 0;
	case NBD_DO_IT:
		if (!lo->file)
			return -EINVAL;
		nbd_pre_doit(lo, dev);
		nbd_do_it(lo);
		nbd_post_doit(lo);
		return lo->harderror;
	case NBD_CLEAR_QUE:
		nbd_clear_que(lo);
		return 0;
#ifdef PARANOIA
	case NBD_PRINT_DEBUG:
		printk(KERN_INFO "NBD device %d: head = %lx, tail = %lx. Global: in %d, out %d\n",
		       dev, (long) lo->head, (long) lo->tail, requests_in, requests_out);
		return 0;
#endif
	case BLKGETSIZE:
		return put_user(nbd_bytesizes[dev] >> 9, (long *) arg);
	}
	return -EINVAL;
}

static int nbd_release(struct inode *inode, struct file *file)
{
	struct nbd_device *lo;
	int dev;

	if (!inode)
		return -ENODEV;
	dev = MINOR(inode->i_rdev);
	if (dev >= MAX_NBD)
		return -ENODEV;
	fsync_dev(inode->i_rdev);
	invalidate_buffers(inode->i_rdev);
	lo = &nbd_dev[dev];
	if (lo->refcnt <= 0)
		printk(KERN_ALERT "nbd_release: refcount(%d) <= 0\n", lo->refcnt);
	lo->refcnt--;
	/* N.B. Doesn't lo->file need an fput?? */
	MOD_DEC_USE_COUNT;
	return 0;
}

static struct file_operations nbd_fops =
{
	NULL,			/* lseek - default */
	block_read,		/* read - general block-dev read */
	block_write,		/* write - general block-dev write */
	NULL,			/* readdir - bad */
	NULL,			/* select */
	nbd_ioctl,		/* ioctl */
	NULL,			/* mmap */
	nbd_open,		/* open */
	NULL,			/* flush */
	nbd_release		/* release */
};

/*
 * And here should be modules and kernel interface 
 *  (Just smiley confuses emacs :-)
 */

#ifdef MODULE
#define nbd_init init_module
#endif

int nbd_init(void)
{
	int i;

	if (sizeof(struct nbd_request) != 28) {
		printk(KERN_CRIT "Sizeof nbd_request needs to be 28 in order to work!\n" );
		return -EIO;
	}

	if (register_blkdev(MAJOR_NR, "nbd", &nbd_fops)) {
		printk("Unable to get major number %d for NBD\n",
		       MAJOR_NR);
		return -EIO;
	}
#ifdef MODULE
	printk("nbd: registered device at major %d\n", MAJOR_NR);
#endif
	blksize_size[MAJOR_NR] = nbd_blksizes;
	blk_size[MAJOR_NR] = nbd_sizes;
	blk_dev[MAJOR_NR].request_fn = do_nbd_request;
	for (i = 0; i < MAX_NBD; i++) {
		memset(nbd_dev + i, 0, sizeof(struct nbd_device));
		nbd_dev[i].magic = LO_MAGIC;
		nbd_dev[i].replysize = NBD_REPLY_SIZE_V0;
		nbd_blksizes[i] = 1024;
		nbd_blksize_bits[i] = 10;
		nbd_bytesizes[i] = 0x7ffffc00; /* 2GB */
		nbd_sizes[i] = nbd_bytesizes[i] >> nbd_blksize_bits[i];
	}
	return 0;
}

#ifdef MODULE
void cleanup_module(void)
{
	if (unregister_blkdev(MAJOR_NR, "nbd") != 0)
		printk("nbd: cleanup_module failed\n");
	else
		printk("nbd: module cleaned up.\n");
}
#endif
